const std = @import("std");

/// Simple ringbuffer-based queue. Capacity will always be a power of two.
pub fn Queue(comptime T: type) type {
    return struct {
        const Self = @This();
        const initial_capacity: usize = 16;

        backing: []T,
        reads: usize,
        writes: usize,
        allocator: std.mem.Allocator,

        pub fn init(allocator: std.mem.Allocator) Self {
            return Self{
                .backing = &[0]T{},
                .reads = 0,
                .writes = 0,
                .allocator = allocator,
            };
        }

        pub fn deinit(self: *Self) void {
            self.clearRetainingCapacity();
            self.allocator.free(self.backing);
        }

        pub fn clearRetainingCapacity(self: *Self) void {
            self.reads = 0;
            self.writes = 0;
        }

        pub fn size(self: *const Self) usize {
            return self.writes - self.reads;
        }

        pub fn capacity(self: *const Self) usize {
            return self.backing.len;
        }

        pub fn ensureTotalCapacity(self: *Self, new_capacity: usize) !void {
            const capacity_sanitized = if (new_capacity > initial_capacity)
                new_capacity
            else
                initial_capacity;

            const new_width = std.math.ceilPowerOfTwoAssert(usize, capacity_sanitized);
            const old_width = self.backing.len;
            if (new_width > old_width) {
                const old_slice = self.backing;
                const new_slice = try self.allocator.alloc(T, new_width);
                const reads = self.reads;
                const len = self.writes - self.reads;
                const mask = old_width -% 1;
                var i: usize = 0;
                while (i < len) : (i += 1) {
                    const j = (reads + i) & mask;
                    new_slice[i] = old_slice[j];
                }

                self.allocator.free(old_slice);
                self.backing = new_slice;
                self.reads = 0;
                self.writes = len;
            }
        }

        pub fn ensureUnusedCapacity(self: *Self, additional_count: usize) !void {
            return self.ensureTotalCapacity(self.size() + additional_count);
        }

        pub fn push(self: *Self, val: T) !void {
            try self.ensureUnusedCapacity(1);
            const mask = self.backing.len - 1;
            const dst = self.writes;
            self.writes += 1;
            self.backing[dst & mask] = val;
        }

        /// Pop pointer to the memory. This may be invalidated on any insert or remove,
        /// so should be copied or discarded before that point.
        pub fn popPtrOrNull(self: *Self) ?*T {
            if (self.size() <= 0) {
                return null;
            }

            // length is assumed to be a power of 2
            const mask = self.backing.len - 1;
            const src = self.reads;
            self.reads += 1;

            // (Same as % self.backing.len because power of 2)
            return &self.backing[src & mask];
        }

        /// Pop pointer to the memory, but panic if null. This may be invalidated on any insert or remove,
        /// so should be copied or discarded before that point.
        pub fn popPtr(self: *Self) *T {
            return self.popPtrOrNull() orelse unreachable;
        }

        pub fn popOrNull(self: *Self) ?T {
            return if (self.popPtrOrNull()) |ptr| ptr.* else null;
        }

        pub fn pop(self: *Self) T {
            return self.popPtr().*;
        }
    };
}

test "general queue usage" {
    var q = Queue(u32).init(std.testing.allocator);
    defer q.deinit();

    try std.testing.expectEqual(@as(?u32, null), q.popOrNull());

    try q.push(45);
    try q.push(42);
    try q.push(2);
    try q.push(0);

    try std.testing.expectEqual(@as(u32, 45), q.pop());
    try std.testing.expectEqual(@as(u32, 42), q.pop());

    try q.push(56);

    var i: u32 = 0;
    while (i < 16) : (i += 1) {
        try q.push(i);
    }

    try std.testing.expectEqual(@as(u32, 2), q.popOrNull() orelse unreachable);
    try std.testing.expectEqual(@as(u32, 0), q.pop());
    try std.testing.expectEqual(@as(u32, 56), q.pop());

    i = 0;
    while (i < 16) : (i += 1) {
        try std.testing.expectEqual(i, q.pop());
    }

    try std.testing.expectEqual(@as(usize, 32), q.capacity());

    try std.testing.expectEqual(@as(?u32, null), q.popOrNull());
    try std.testing.expectEqual(@as(?u32, null), q.popOrNull());
}
