const std = @import("std");
const assert = std.debug.assert;
const vk = @import("vulkan");
const vma = @import("vma");
const time = @import("../../common/time.zig");

const ProfileMark = @import("../../common/profiler.zig").ProfileMark;
const Device = @import("device.zig").Device;
const Swapchain = @import("swapchain.zig").Swapchain;
const settings = @import("settings.zig");
const Command = @import("Command.zig");
const SubmitId = @import("submit_id.zig").SubmitId;
const memory = @import("memory.zig");
const queues = @import("queues.zig");

const Buffer = @This();

state: State,
handle: vk.Buffer,
allocation: vma.Allocation,
size: usize,

pub fn init(size: usize, usage: vk.BufferUsageFlags, mem_usage: vma.MemoryUsage) !Buffer {
    var buffer = std.mem.zeroes(Buffer);
    try memory.bufferNew(&buffer, size, usage, mem_usage);
    return buffer;
}

pub fn release(self: *Buffer, device: *const Device) !void {
    if (self.handle != .null_handle) {
        const id = try self.getSubmit();
        const frame_count = time.getFrameCount();
        _ = try memory.Releasable.init(
            device,
            frame_count,
            id,
            .{
                .buffer = .{
                    .handle = self.handle,
                    .allocation = self.allocation,
                },
            },
        );
    }
    self.* = std.mem.zeroes(Buffer);
}

const pm_reserve = ProfileMark.init("Buffer.reserve");
pub fn reserve(self: *const Buffer, size: usize, buffer_usage: vk.BufferUsageFlags, mem_usage: vma.MemUsage) !void {
    pm_reserve.begin();
    defer pm_reserve.end();

    const old_size = self.size;
    assert(old_size >= 0);
    if (old_size < size) {
        self.release();
        const new_size = if (size > old_size * 2) size else old_size * 2;
        try self.init(new_size, buffer_usage, mem_usage);
    }
}

pub fn mapRead(self: *const Buffer, device: *const Device) !anyopaque {
    if (!self.state.stage.host_bit || !self.state.access.host_read_bit) {
        // kara claims this sucks but she's lazy.
        var cmd = try Command.Buffer.get(.graphics);
        _ = try State.hostRead(device, cmd, self);
        const submit_id = try cmd.submit(device, .null_handle, .{}, .null_handle);
        try submit_id.wait(device);
        cmd = try Command.Buffer.get(.graphics);
        assert(cmd != .null_handle);
    }

    assert(self.state.stage.host_bit);
    assert(self.state.access.host_read_bit);
    return memory.map(self.allocation);
}

pub fn unmapRead(self: *const Buffer, device: *const Device) void {
    assert(self.state.stage.host_bit);
    assert(self.state.access.host_read_bit);
    memory.unmap(device, self.allocation);
}

pub fn mapWrite(self: *Buffer, device: *const Device) ![*]u8 {
    if (self.state.stage.toInt() == 0) {
        // newly created buffer
        self.state.stage = .{ .host_bit = true };
        self.state.access = .{ .host_write_bit = true };
    }

    if (!self.state.stage.host_bit or !self.state.access.host_write_bit) {
        // kara claims this sucks but she's lazy
        var cmd = try Command.Buffer.get(.graphics, device);
        _ = try State.hostWrite(device, cmd, self);
        const submit_id = try cmd.submit(device, null, .{}, null);
        try submit_id.wait(device);
        cmd = try Command.Buffer.get(.graphics, device);
        assert(cmd.handle != .null_handle);
    }

    assert(self.state.stage.host_bit);
    assert(self.state.access.host_write_bit);
    return try memory.map(self.allocation);
}

pub fn unmapWrite(self: *const Buffer) void {
    assert(self.state.stage.host_bit);
    assert(self.state.access.host_write_bit);
    memory.unmap(self.allocation);
    memory.flush(self.allocation);
}

const pm_write = ProfileMark.init("Buffer.write");
pub fn write(self: *Buffer, device: *const Device, src: []const u8) !void {
    pm_write.begin();
    defer pm_write.end();

    assert(src.len <= self.size);
    var dst = try self.mapWrite(device);
    defer self.unmapWrite();
    @memcpy(dst, src.ptr, src.len);
}

const pm_read = ProfileMark.init("Buffer.read");
pub fn read(self: *const Buffer, device: *const Device, dst: *[]u8) !void {
    pm_read.begin();
    defer pm_read.end();

    assert(dst.len == self.size);
    const src = try self.mapRead(device);
    defer self.unmapRead(device);
    @memcpy(dst, src, dst.len);
}

pub fn getSubmit(self: *const Buffer) !SubmitId {
    if (self.handle != .null_handle) {
        const id = if (self.state.stage.toInt() != 0) SubmitId{
            .counter = self.state.cmd_id,
            .queue_id = self.state.owner,
            .valid = true,
        } else Command.getHeadSubmit(self.state.owner);

        assert(id.valid);
        return id;
    }

    return error.InvalidBufferHandle;
}

pub const State = struct {
    owner: queues.QueueId,
    cmd_id: u32 = 0,
    stage: vk.PipelineStageFlags,
    access: vk.AccessFlags,

    pub fn hostRead(device: *const Device, cmdbuf: *const Command.Buffer, buf: *Buffer) !bool {
        assert(cmdbuf.handle != .null_handle);
        assert(cmdbuf.xfer);
        const state = State{
            .owner = cmdbuf.queue_id,
            .stage = .{ .host_bit = true },
            .access = .{ .host_read_bit = true },
        };
        return try state.handle(device, buf);
    }

    pub fn hostWrite(device: *const Device, cmdbuf: *const Command.Buffer, buf: *Buffer) !bool {
        assert(cmdbuf.handle != .null_handle);
        assert(cmdbuf.xfer);
        const state = State{
            .owner = cmdbuf.queue_id,
            .stage = .{ .host_bit = true },
            .access = .{ .host_write_bit = true },
        };
        return try state.handle(device, buf);
    }

    pub fn transferSrc(device: *const Device, cmdbuf: *const Command.Buffer, buf: *Buffer) !bool {
        assert(cmdbuf.handle != .null_handle);
        assert(cmdbuf.xfer);
        const state = State{
            .owner = cmdbuf.queue_id,
            .stage = .{ .transfer_bit = true },
            .access = .{ .transfer_read_bit = true },
        };
        return try state.handle(device, buf);
    }

    pub fn transferDst(device: *const Device, cmdbuf: *const Command.Buffer, buf: *Buffer) !bool {
        assert(cmdbuf.handle != .null_handle);
        assert(cmdbuf.xfer);
        const state = State{
            .owner = cmdbuf.queue_id,
            .stage = .{ .transfer_bit = true },
            .access = .{ .transfer_write_bit = true },
        };
        return try state.handle(device, buf);
    }

    pub fn uniformBuffer(device: *const Device, cmdbuf: *const Command.Buffer, buf: *Buffer) !bool {
        assert(cmdbuf.handle != .null_handle);
        assert(cmdbuf.gfx or cmdbuf.comp);
        const state = State{
            .owner = cmdbuf.queue_id,
            .stage = .{ .vertex_shader_bit = true, .fragment_shader_bit = true, .compute_shader_bit = true },
            .access = .{ .uniform_read_bit = true },
        };
        return try state.handle(device, buf);
    }

    pub fn indirectDraw(device: *const Device, cmdbuf: *const Command.Buffer, buf: *Buffer) !bool {
        assert(cmdbuf.handle != .null_handle);
        assert(cmdbuf.gfx);
        const state = State{
            .owner = cmdbuf.queue_id,
            .stage = .{ .draw_indirect_bit = true },
            .access = .{ .indirect_command_read_bit = true },
        };
        return try state.handle(device, buf);
    }

    pub fn indirectDispatch(device: *const Device, cmdbuf: *const Command.Buffer, buf: *Buffer) !bool {
        assert(cmdbuf.handle != .null_handle);
        assert(cmdbuf.comp);
        const state = State{
            .owner = cmdbuf.queue_id,
            // kara was not sure if draw indirect stage is needed here or not
            .stage = .{ .draw_indirect_bit = true, .compute_shader_bit = true },
            .access = .{ .indirect_command_read_bit = true },
        };
        return try state.handle(device, buf);
    }

    pub fn vertexBuffer(device: *const Device, cmdbuf: *const Command.Buffer, buf: *Buffer) !bool {
        assert(cmdbuf.handle != .null_handle);
        assert(cmdbuf.gfx);
        const state = State{
            .owner = cmdbuf.queue_id,
            .stage = .{ .vertex_input_bit = true },
            .access = .{ .vertex_attribute_read_bit = true },
        };
        return try state.handle(device, buf);
    }

    pub fn indexBuffer(device: *const Device, cmdbuf: *const Command.Buffer, buf: *Buffer) !bool {
        assert(cmdbuf.handle != .null_handle);
        assert(cmdbuf.gfx);
        const state = State{
            .owner = cmdbuf.queue_id,
            .stage = .{ .vertex_input_bit = true },
            .access = .{ .index_read_bit = true },
        };
        return try state.handle(device, buf);
    }

    pub fn fragLoad(device: *const Device, cmdbuf: *const Command.Buffer, buf: *Buffer) !bool {
        assert(cmdbuf.handle != .null_handle);
        assert(cmdbuf.gfx);
        const state = State{
            .owner = cmdbuf.queue_id,
            .stage = .{ .fragment_shader_bit = true },
            .access = .{ .shader_read_bit = true },
        };
        return try state.handle(device, buf);
    }

    pub fn fragStore(device: *const Device, cmdbuf: *const Command.Buffer, buf: *Buffer) !bool {
        assert(cmdbuf.handle != .null_handle);
        assert(cmdbuf.gfx);
        const state = State{
            .owner = cmdbuf.queue_id,
            .stage = .{ .fragment_shader_bit = true },
            .access = .{ .shader_write_bit = true },
        };
        return try state.handle(device, buf);
    }

    pub fn fragLoadStore(device: *const Device, cmdbuf: *const Command.Buffer, buf: *Buffer) !bool {
        assert(cmdbuf.handle != .null_handle);
        assert(cmdbuf.gfx);
        const state = State{
            .owner = cmdbuf.queue_id,
            .stage = .{ .fragment_shader_bit = true },
            .access = .{ .shader_read_bit = true, .shader_write_bit = true },
        };
        return try state.handle(device, buf);
    }

    pub fn computeLoad(device: *const Device, cmdbuf: *const Command.Buffer, buf: *Buffer) !bool {
        assert(cmdbuf.handle != .null_handle);
        assert(cmdbuf.comp);
        const state = State{
            .owner = cmdbuf.queue_id,
            .stage = .{ .compute_shader_bit = true },
            .access = .{ .shader_read_bit = true },
        };
        return try state.handle(device, buf);
    }

    pub fn computeStore(device: *const Device, cmdbuf: *const Command.Buffer, buf: *Buffer) !bool {
        assert(cmdbuf.handle != .null_handle);
        assert(cmdbuf.comp);
        const state = State{
            .owner = cmdbuf.queue_id,
            .stage = .{ .compute_shader_bit = true },
            .access = .{ .shader_write_bit = true },
        };
        return try state.handle(device, buf);
    }

    pub fn computeLoadStore(device: *const Device, cmdbuf: *const Command.Buffer, buf: *Buffer) !bool {
        assert(cmdbuf.handle != .null_handle);
        assert(cmdbuf.comp);
        const state = State{
            .owner = cmdbuf.queue_id,
            .stage = .{ .compute_shader_bit = true },
            .access = .{ .shader_read_bit = true, .shader_write_bit = true },
        };
        return try state.handle(device, buf);
    }
    fn handle(self: *const State, device: *const Device, buf: *Buffer) !bool {
        assert(buf.handle != .null_handle);
        assert(self.access.toInt() != 0);
        assert(self.stage.toInt() != 0);

        var inserted_barrier = false;
        const prev = &buf.state;
        const prev_queue = queues.get(prev.owner);
        const next_queue = queues.get(self.owner);

        assert(self.stage.intersect(next_queue.stage_mask).toInt() == self.stage.toInt());
        assert(self.access.intersect(next_queue.access_mask).toInt() == self.access.toInt());

        var new_resource = false;
        if (prev.stage.toInt() == 0) {
            prev.stage = .{ .top_of_pipe_bit = true };
            new_resource = true;
        }

        if (prev_queue != next_queue) {
            // queue ownership transfer
            var src_cmd = try Command.Buffer.get(prev.owner, device);
            var dst_cmd = try Command.Buffer.get(self.owner, device);
            if (src_cmd.queue_transfer_dst or dst_cmd.queue_transfer_src) {
                try Command.Buffer.flushQueueTransfers(device);
            }

            const barrier = vk.BufferMemoryBarrier{
                .src_access_mask = prev.access,
                .dst_access_mask = self.access,
                .src_queue_family_index = prev_queue.family,
                .dst_queue_family_index = next_queue.family,
                .buffer = buf.handle,
                .offset = 0,
                .size = vk.WHOLE_SIZE,
            };

            assert(!src_cmd.queue_transfer_dst);
            src_cmd.queue_transfer_src = true;
            device.dispatch.cmdPipelineBarrier(src_cmd.handle, prev.stage, self.stage, .{}, 0, undefined, 1, @ptrCast([*]const vk.BufferMemoryBarrier, &barrier), 0, undefined);
            assert(!dst_cmd.queue_transfer_src);
            dst_cmd.queue_transfer_dst = true;
            device.dispatch.cmdPipelineBarrier(dst_cmd.handle, prev.stage, self.stage, .{}, 0, undefined, 1, @ptrCast([*]const vk.BufferMemoryBarrier, &barrier), 0, undefined);
            prev.* = self.*;
            inserted_barrier = true;
        } else {
            var need_barrier = false;
            const host_only = prev.stage.host_bit and self.stage.host_bit;
            if (!host_only) {
                const src_read = prev.access.intersect(Command.read_access).toInt() != 0;
                const src_write = prev.access.intersect(Command.write_access).toInt() != 0;
                const dst_read = self.access.intersect(Command.read_access).toInt() != 0;
                const dst_write = self.access.intersect(Command.write_access).toInt() != 0;
                const read_after_write = src_write and dst_read;
                const write_after_read = src_read and dst_write;
                const write_after_write = src_write and dst_write;
                need_barrier = read_after_write or write_after_read or write_after_write or new_resource;
            }

            if (need_barrier) {
                // data hazard, insert barrier
                const cmd = try Command.Buffer.get(self.owner, device);
                assert(!cmd.in_render_pass);
                const barrier = vk.BufferMemoryBarrier{
                    .src_access_mask = prev.access,
                    .dst_access_mask = self.access,
                    .src_queue_family_index = vk.QUEUE_FAMILY_IGNORED,
                    .dst_queue_family_index = vk.QUEUE_FAMILY_IGNORED,
                    .buffer = buf.handle,
                    .offset = 0,
                    .size = vk.WHOLE_SIZE,
                };
                device.dispatch.cmdPipelineBarrier(cmd.handle, prev.stage, self.stage, .{}, 0, undefined, 1, @ptrCast([*]const vk.BufferMemoryBarrier, &barrier), 0, undefined);
                prev.* = self.*;
                inserted_barrier = true;
            } else {
                // no data hazard, append usage state
                prev.stage = prev.stage.merge(self.stage);
                prev.access = prev.access.merge(self.access);
            }
        }

        (try Command.Buffer.get(self.owner, device)).touchBuffer(buf);
        return inserted_barrier;
    }
};

pub const Set = struct {
    frames: [settings.resource_sets]Buffer,

    pub fn init(self: *Set, device: *const Device, size: usize, usage: vk.BufferUsageFlags, mem_usage: vma.MemoryUsage) !void {
        self.* = std.mem.zeroes(Set);
        errdefer self.release(device);

        for (self.frames) |*frame| {
            frame.* = try Buffer.init(size, usage, mem_usage);
        }
    }

    pub fn release(self: *Set, device: *const Device) void {
        for (self.frames) |*frame| {
            frame.release(device) catch {};
        }
        self.* = std.mem.zeroes(Set);
    }

    pub fn reserve(self: *const Set, device: *const Device, swapchain: *const Swapchain, size: usize, usage: vk.BufferUsageFlags, mem_usage: vma.MemoryUsage) !void {
        return try self.current(swapchain).reserve(device, size, usage, mem_usage);
    }

    pub fn write(self: *Set, device: *const Device, swapchain: *const Swapchain, src: []const u8) !void {
        try self.current(swapchain).write(device, src);
    }

    pub fn read(self: *const Set, device: *const Device, swapchain: *const Swapchain, dst: *[]u8) !void {
        try self.current(swapchain).read(device, dst);
    }

    pub fn current(self: *Set, swapchain: *const Swapchain) *Buffer {
        const sync_index = swapchain.getSyncIndex();
        assert(sync_index < self.frames.len);
        return &self.frames[sync_index];
    }

    pub fn next(self: *const Set, swapchain: *const Swapchain) *const Buffer {
        const sync_index = swapchain.getNextSyncIndex();
        assert(sync_index < self.frames.len);
        return &self.frames[sync_index];
    }

    pub fn prev(self: *const Set, swapchain: *const Swapchain) *const Buffer {
        const sync_index = swapchain.getPrevSyncIndex();
        assert(sync_index < self.frames.len);
        return &self.frames[sync_index];
    }
};
