const std = @import("std");
const assert = std.debug.assert;
const vk = @import("vulkan");

const Device = @import("device.zig").Device;
const Bindings = @import("Bindings.zig");
const pipeline = @import("pipeline.zig");

const Self = @This();

pipeline: vk.Pipeline,
layout: vk.PipelineLayout,
bind_point: vk.PipelineBindPoint,
stage_flags: vk.ShaderStageFlags,
push_constant_bytes: u32,

pub const VertexLayout = struct {
    bindings: []const vk.VertexInputBindingDescription,
    attributes: []const vk.VertexInputAttributeDescription,
};

pub const BlendState = struct {
    color_write_mask: vk.ColorComponentFlags = .{},
    blend_enable: bool = false,
    src_color_blend_factor: vk.BlendFactor = .zero,
    dst_color_blend_factor: vk.BlendFactor = .zero,
    color_blend_op: vk.BlendOp,
    src_alpha_blend_factor: vk.BlendFactor = .zero,
    dst_alpha_blend_factor: vk.BlendFactor = .zero,
    alpha_blend_op: vk.BlendOp,
};

pub const FixedFuncs = struct {
    viewport: vk.Viewport,
    scissor: vk.Rect2D,
    topology: vk.PrimitiveTopology,
    polygon_mode: vk.PolygonMode,
    cull_mode: vk.CullModeFlags,
    front_face: vk.FrontFace,
    depth_compare_op: vk.CompareOp,
    scissor_on: bool,
    depth_clamp: bool,
    depth_test_enable: bool,
    depth_write_enable: bool,
    attachments: ?[]const BlendState = null,
};

pub const Description = struct {
    // Graphics and compute
    push_constant_bytes: u32,
    shaders: []const vk.PipelineShaderStageCreateInfo,
    // Graphics only
    render_pass: vk.RenderPass,
    subpass: ?u32,
    vert_layout: VertexLayout,
    fixed_funcs: FixedFuncs,
};

pub fn init(self: *Self, device: *const Device, desc: *const Description) !void {
    self.* = std.mem.zeroes(Self);

    var stage_flags = vk.ShaderStageFlags{};
    for (desc.shaders) |shader| {
        stage_flags = stage_flags.merge(shader.stage);
    }

    var bind_point = vk.PipelineBindPoint.graphics;
    if (stage_flags.compute_bit) {
        bind_point = .compute;
    }

    const k_rt_stages = vk.ShaderStageFlags{
        .raygen_bit_khr = true,
        .any_hit_bit_khr = true,
        .closest_hit_bit_khr = true,
        .miss_bit_khr = true,
        .intersection_bit_khr = true,
        .callable_bit_khr = true,
    };

    if (stage_flags.intersect(k_rt_stages).toInt() != 0) {
        bind_point = .ray_tracing_khr;
    }

    const set_layouts = [_]vk.DescriptorSetLayout{
        Bindings.getSetLayout(),
    };

    const ranges = [_]vk.PushConstantRange{
        .{
            .stage_flags = stage_flags,
            .size = desc.push_constant_bytes,
            .offset = 0,
        },
    };

    self.bind_point = bind_point;
    self.stage_flags = stage_flags;
    self.push_constant_bytes = desc.push_constant_bytes;
    assert(desc.push_constant_bytes <= 256);

    errdefer self.deinit(device);
    self.layout = try pipeline.Layout.init(device, &set_layouts, &ranges);

    switch (bind_point) {
        .compute => {
            assert(desc.shaders.len == 1);
            self.pipeline = try pipeline.newComp(device, self.layout, &desc.shaders[0]);
        },
        .graphics => {
            assert(desc.shaders.len == 2);
            self.pipeline = try pipeline.newGfx(device, &desc.fixed_funcs, &desc.vert_layout, self.layout, desc.render_pass, desc.subpass, desc.shaders);
        },
        else => assert(false),
    }
}

pub fn deinit(self: *Self, device: *const Device) void {
    pipeline.Layout.deinit(device, self.layout);
    pipeline.deinit(device, self.pipeline);
    self.* = std.mem.zeroes(Self);
}
