const std = @import("std");
const vk = @import("vulkan");

const profiler = @import("../../common/profiler.zig");
const Device = @import("device.zig").Device;

// TODO memory
var gpa = std.heap.GeneralPurposeAllocator(.{}){};
const allocator = gpa.allocator();

var s_cache = std.AutoHashMap(Description, vk.RenderPass).init(allocator);

// https://www.khronos.org/registry/vulkan/specs/1.2-extensions/html/vkspec.html#renderpass-compatibility
// compatibility:
// refs: matching format and sample count, or both VK_ATTACHMENT_UNUSED, or both NULL
// arrays of refs: treats missing refs as VK_ATTACHMENT_UNUSED
const AttachmentState = packed struct {
    format: vk.Format, // must match
    layout: vk.ImageLayout, // can vary
    load: vk.AttachmentLoadOp, // can vary
    store: vk.AttachmentStoreOp, // can vary
};

pub const Description = struct {
    attachments: [8]AttachmentState = std.mem.zeroes([8]AttachmentState),
    src_access_mask: vk.AccessFlags,
    dst_access_mask: vk.AccessFlags,
    src_stage_mask: vk.PipelineStageFlags,
    dst_stage_mask: vk.PipelineStageFlags,
};

const pm_get = profiler.ProfileMark.init("renderpass.get");
pub fn get(device: *const Device, desc: Description) !vk.RenderPass {
    pm_get.begin();
    defer pm_get.end();

    const result = try s_cache.getOrPut(desc);
    if (result.found_existing) {
        return result.value_ptr.*;
    }

    var attachments = std.mem.zeroes([8]vk.AttachmentDescription);
    var refs = std.mem.zeroes([attachments.len]vk.AttachmentReference);
    var attachment_count: u32 = 0;
    var ref_count: usize = 0;
    const format0 = desc.attachments[0].format;
    const zero_is_depth = @enumToInt(format0) >= @enumToInt(vk.Format.d16_unorm) and @enumToInt(format0) <= @enumToInt(vk.Format.d32_sfloat_s8_uint);

    for (desc.attachments) |*src, i| {
        attachments[i].load_op = .dont_care;
        attachments[i].store_op = .dont_care;
        if (src.format != .@"undefined") {
            std.debug.assert(attachment_count < attachments.len);
            var dst = &attachments[attachment_count];
            dst.format = src.format;
            dst.initial_layout = src.layout;
            dst.final_layout = src.layout;
            dst.samples = .{ .@"1_bit" = true };
            dst.stencil_load_op = .dont_care;
            dst.stencil_store_op = .dont_care;
            dst.load_op = src.load;
            dst.store_op = src.store;
            refs[attachment_count].attachment = attachment_count;
            refs[attachment_count].layout = src.layout;

            if (dst.load_op != .dont_care or dst.store_op != .dont_care) {
                std.debug.assert(ref_count < refs.len);
                refs[attachment_count].attachment = attachment_count;
                refs[attachment_count].layout = src.layout;
            }

            attachment_count += 1;
        }
    }

    std.debug.assert(attachment_count <= attachments.len);
    std.debug.assert(attachment_count <= refs.len);
    const color_ref_count = if (zero_is_depth) attachment_count - 1 else attachment_count;
    const depth_ref_count = if (zero_is_depth) @as(u32, 1) else 0;

    const subpasses = [_]vk.SubpassDescription{.{
        .flags = .{},
        .input_attachment_count = 0,
        .p_input_attachments = undefined,
        .p_resolve_attachments = null,
        .preserve_attachment_count = 0,
        .p_preserve_attachments = undefined,
        .pipeline_bind_point = .graphics,
        .color_attachment_count = color_ref_count,
        .p_color_attachments = if (color_ref_count > 0) @ptrCast([*]const vk.AttachmentReference, &refs[depth_ref_count]) else undefined,
        .p_depth_stencil_attachment = if (depth_ref_count > 0) @as(?*const vk.AttachmentReference, &refs[0]) else null,
    }};

    const dependencies = [_]vk.SubpassDependency{.{
        .dependency_flags = .{},
        .src_subpass = vk.SUBPASS_EXTERNAL,
        .dst_subpass = 0,
        .src_stage_mask = desc.src_stage_mask,
        .dst_stage_mask = desc.dst_stage_mask,
        .src_access_mask = desc.src_access_mask,
        .dst_access_mask = desc.dst_access_mask,
    }};

    const info = vk.RenderPassCreateInfo{
        .flags = .{},
        .attachment_count = attachment_count,
        .p_attachments = &attachments,
        .subpass_count = subpasses.len,
        .p_subpasses = &subpasses,
        .dependency_count = dependencies.len,
        .p_dependencies = &dependencies,
    };
    const handle = try device.dispatch.createRenderPass(device.handle, &info, null);

    result.value_ptr.* = handle;
    return result.value_ptr.*;
}
