const assert = @import("std").debug.assert;
const vk = @import("vulkan");

const ProfileMark = @import("../../common/profiler.zig").ProfileMark;
const settings = @import("settings.zig");
const Device = @import("device.zig").Device;

const pm_writeimagetable = ProfileMark.init("descriptor.writeImageTable");
pub fn writeImageTable(device: *const Device, set: vk.DescriptorSet, binding: i32, desc_type: vk.DescriptorType, bindings: []const vk.DescriptorImageInfo) void {
    pm_writeimagetable.begin();
    defer pm_writeimagetable.end();

    device.dispatch.updateDescriptorSets(device.handle, 1, @ptrCast([*]const vk.WriteDescriptorSet, &.{
        .descriptor_type = desc_type,
        .dst_set = set,
        .dst_binding = binding,
        .dst_array_element = 0,
        .descriptor_count = bindings.len,
        .p_image_info = bindings.ptr,
    }), 0, null);
}

const pm_writebuffertable = ProfileMark.init("descriptor.writeBufferTable");
pub fn writeBufferTable(device: *const Device, set: vk.DescriptorSet, binding: i32, desc_type: vk.DescriptorType, bindings: []const vk.DescriptorBufferInfo) void {
    assert(set != .null_handle);
    assert(@enumToInt(desc_type) >= @enumToInt(vk.DescriptorType.uniform_texel_buffer));
    assert(@enumToInt(desc_type) <= @enumToInt(vk.DescriptorType.storage_buffer_dynamic));

    pm_writebuffertable.begin();
    defer pm_writebuffertable.end();

    device.dispatch.updateDescriptorSets(device.handle, 1, @ptrCast([*]vk.WriteDescriptorSet, &.{
        .descriptor_type = type,
        .dst_set = set,
        .dst_binding = binding,
        .dst_array_element = 0,
        .descriptor_count = bindings.len,
        .p_buffer_info = bindings.ptr,
    }), 0, null);
}

pub const Set = struct {
    pub fn init(device: *const Device, pool: vk.DescriptorPool, layout: vk.DescriptorSetLayout) !vk.DescriptorSet {
        assert(pool != .null_handle);
        assert(layout != .null_handle);
        var set: vk.DescriptorSet = undefined;
        _ = try device.dispatch.allocateDescriptorSets(device.handle, &.{
            .descriptor_pool = pool,
            .descriptor_set_count = 1,
            .p_set_layouts = @ptrCast([*]const vk.DescriptorSetLayout, &layout),
        }, @ptrCast([*]vk.DescriptorSet, &set));
        return set;
    }

    pub fn deinit(device: *const Device, pool: vk.DescriptorPool, set: vk.DescriptorSet) void {
        assert(pool != .null_handle);
        if (set != .null_handle) {
            device.dispatch.freeDescriptorSets(device.handle, pool, 1, @ptrCast([*]const vk.DescriptorSet, &set)) catch unreachable;
        }
    }

    pub const Layout = struct {
        pub fn init(device: *const Device, bindings: []const vk.DescriptorSetLayoutBinding, flags: vk.DescriptorSetLayoutCreateFlags) !vk.DescriptorSetLayout {
            return try device.dispatch.createDescriptorSetLayout(device.handle, &.{
                .flags = flags,
                .binding_count = @intCast(u32, bindings.len),
                .p_bindings = bindings.ptr,
            }, null);
        }

        pub fn deinit(device: *const Device, handle: vk.DescriptorSetLayout) void {
            if (handle != .null_handle) {
                return device.dispatch.destroyDescriptorSetLayout(device.handle, handle, null);
            }
        }
    };
};

pub const Pool = struct {
    pub const sizes = [_]vk.DescriptorPoolSize{
        .{
            .@"type" = .uniform_buffer,
            .descriptor_count = settings.uniform_buffer_count,
        },
        .{
            .@"type" = .storage_buffer,
            .descriptor_count = settings.storage_buffer_count,
        },
        .{
            .@"type" = .storage_image,
            .descriptor_count = settings.storage_image_count,
        },
        .{
            .@"type" = .combined_image_sampler,
            .descriptor_count = settings.combined_image_sampler_count,
        },
    };

    pub fn init(device: *const Device, max_sets: u32, pool_sizes: []const vk.DescriptorPoolSize) !vk.DescriptorPool {
        assert(max_sets > 0);
        return try device.dispatch.createDescriptorPool(device.handle, &.{
            .flags = .{ .free_descriptor_set_bit = true },
            .max_sets = max_sets,
            .pool_size_count = @intCast(u32, pool_sizes.len),
            .p_pool_sizes = pool_sizes.ptr,
        }, null);
    }

    pub fn deinit(device: *const Device, pool: vk.DescriptorPool) void {
        if (pool != .null_handle) {
            device.dispatch.destroyDescriptorPool(device.handle, pool, null);
        }
    }

    const pm_pool_reset = ProfileMark.init("descriptor.Pool.reset");
    pub fn reset(device: *const Device, pool: vk.DescriptorPool) !void {
        pm_pool_reset.begin();
        defer pm_pool_reset.end();
        assert(pool != .null_handle);
        try device.dispatch.resetDescriptorPool(device.handle, pool, .{});
    }
};
