const std = @import("std");
const vk = @import("vulkan");
const glfw = @import("glfw");

const settings = @import("settings.zig");
const Base = @import("Base.zig");
const Extensions = @import("Extensions.zig");
const enabled_layers = @import("layers.zig").enabled;

const DebugMessenger = @import("DebugMessenger.zig");

// TODO temp allocator
var gpa = std.heap.GeneralPurposeAllocator(.{}){};
const allocator = gpa.allocator();

// TODO: filter by settings!
const InstanceDispatch = vk.InstanceWrapper(.{
    .destroyInstance = true,
    .createDevice = true,
    .destroySurfaceKHR = true,
    .enumeratePhysicalDevices = true,
    .createDebugUtilsMessengerEXT = settings.messenger_on,
    .destroyDebugUtilsMessengerEXT = settings.messenger_on,
    .getPhysicalDeviceProperties2 = true,
    .getPhysicalDeviceFeatures2 = true,
    .enumerateDeviceExtensionProperties = true,
    .getPhysicalDeviceSurfaceFormatsKHR = true,
    .getPhysicalDeviceSurfacePresentModesKHR = true,
    .getPhysicalDeviceSurfaceCapabilitiesKHR = true,
    .getPhysicalDeviceQueueFamilyProperties = true,
    .getPhysicalDeviceSurfaceSupportKHR = true,
    .getPhysicalDeviceMemoryProperties = true,
    .getDeviceProcAddr = true,
});

pub const Instance = struct {
    const Self = @This();

    dispatch: InstanceDispatch,
    handle: vk.Instance,
    messenger: DebugMessenger = undefined,

    layers: std.ArrayList([*:0]const u8),
    extensions: std.ArrayList([*:0]const u8),

    pub fn init() !Self {
        // TODO: do we absolutely need volk rn? Don't think so? volk init here
        var avail_layers = try getAvailableLayers();
        defer allocator.free(avail_layers);
        var avail_exts = try getAvailableExtensions();
        defer allocator.free(avail_exts);

        const layers_to_load = try getLayers(avail_layers);
        const extensions = try getExtensions(avail_exts);

        const app_info = vk.ApplicationInfo{
            .p_application_name = settings.app_name,
            .application_version = vk.makeApiVersion(1, 0, 0, 0),
            .p_engine_name = settings.engine_name,
            .engine_version = vk.makeApiVersion(1, 0, 0, 0),
            .api_version = vk.API_VERSION_1_2,
        };

        const handle = try Base.dispatch.createInstance(&.{
            .flags = .{},
            .p_application_info = &app_info,
            .enabled_layer_count = @intCast(u32, layers_to_load.items.len),
            .pp_enabled_layer_names = layers_to_load.items.ptr,
            .enabled_extension_count = @intCast(u32, extensions.items.len),
            .pp_enabled_extension_names = extensions.items.ptr,
        }, null);

        const dispatch = try InstanceDispatch.load(handle, Base.vk_proc);
        errdefer dispatch.destroyInstance(handle, null);

        var self = Self{
            .dispatch = dispatch,
            .layers = layers_to_load,
            .extensions = extensions,
            .handle = handle,
        };
        self.messenger = if (settings.messenger_on) try DebugMessenger.init(&self) else undefined;
        return self;
    }

    pub fn deinit(self: Self) void {
        self.dispatch.destroyInstance(self.handle, null);

        self.layers.deinit();
        self.extensions.deinit();
        self.messenger.deinit(&self);
    }

    /// list all available layers
    fn getAvailableLayers() ![]vk.LayerProperties {
        var count: u32 = undefined;
        _ = try Base.dispatch.enumerateInstanceLayerProperties(&count, null);
        const list = try allocator.alloc(vk.LayerProperties, count);
        _ = try Base.dispatch.enumerateInstanceLayerProperties(&count, list.ptr);

        // TODO log
        std.debug.print("{} available instance layers\n", .{count});
        for (list) |layer| {
            const layer_name = std.mem.sliceTo(&layer.layer_name, 0);
            std.debug.print("\t{s}\n", .{layer_name});
        }

        return list;
    }

    fn getLayers(avail_layers: []vk.LayerProperties) !std.ArrayList([*:0]const u8) {
        var list = std.ArrayList([*:0]const u8).init(allocator);

        for (enabled_layers) |layer| {
            const layer_name = std.mem.sliceTo(layer, 0);
            for (avail_layers) |avail_layer| {
                const avail_layer_name = std.mem.sliceTo(&avail_layer.layer_name, 0);
                if (std.mem.eql(u8, layer_name, avail_layer_name)) {
                    try list.append(layer);
                }
            }
        }

        return list;
    }

    /// list all available extensions
    fn getAvailableExtensions() ![]vk.ExtensionProperties {
        var count: u32 = undefined;
        _ = try Base.dispatch.enumerateInstanceExtensionProperties(null, &count, null);
        const list = try allocator.alloc(vk.ExtensionProperties, count);
        _ = try Base.dispatch.enumerateInstanceExtensionProperties(null, &count, list.ptr);

        // TODO log
        std.debug.print("{} available instance extensions\n", .{count});
        for (list) |ext| {
            const extension_name = std.mem.sliceTo(&ext.extension_name, 0);
            std.debug.print("\t{s}\n", .{extension_name});
        }

        return list;
    }

    fn getExtensions(avail_exts: []vk.ExtensionProperties) !std.ArrayList([*:0]const u8) {
        var list = std.ArrayList([*:0]const u8).init(allocator);

        for (try glfw.getRequiredInstanceExtensions()) |ext| {
            const extension_name = std.mem.sliceTo(ext, 0);
            for (avail_exts) |avail_ext| {
                const avail_ext_name = std.mem.sliceTo(&avail_ext.extension_name, 0);
                if (std.mem.eql(u8, extension_name, avail_ext_name)) {
                    try list.append(extension_name);
                }
            }
        }

        const instExts = try Extensions.Instance.toArray(allocator);
        defer allocator.free(instExts);
        for (instExts) |ext| {
            const extension_name = std.mem.sliceTo(ext, 0);
            for (avail_exts) |avail_ext| {
                const avail_ext_name = std.mem.sliceTo(&avail_ext.extension_name, 0);
                if (std.mem.eql(u8, extension_name, avail_ext_name)) {
                    try list.append(ext);
                }
            }
        }

        return list;
    }
};
