const std = @import("std");
const assert = std.debug.assert;

const vk = @import("vulkan");
const Device = @import("device.zig").Device;
const DeviceDispatch = @import("device.zig").DeviceDispatch;

const Renderer = @import("Renderer.zig");

pub const Semaphore = struct {
    const Self = @This();

    handle: vk.Semaphore,

    pub fn init(device: *const Device) !Self {
        const handle = try device.dispatch.createSemaphore(device.handle, &.{
            .flags = .{},
        }, null);

        return Self{
            .handle = handle,
        };
    }

    pub fn deinit(self: Self, device: *const Device) void {
        device.dispatch.destroySemaphore(device.handle, self.handle, null);
    }
};

pub const Fence = struct {
    const Self = @This();

    pub const State = enum(i32) {
        signaled = @enumToInt(vk.Result.success),
        unsignaled = @enumToInt(vk.Result.not_ready),
    };

    handle: vk.Fence,

    pub const InitError = DeviceDispatch.CreateFenceError;
    pub fn init(device: *const Device, signaled: bool) !Self {
        const handle = try device.dispatch.createFence(device.handle, &.{ .flags = .{
            .signaled_bit = if (signaled) true else false,
        } }, null);

        return Self{
            .handle = handle,
        };
    }

    pub fn deinit(self: *const Self, device: *const Device) void {
        device.dispatch.destroyFence(device.handle, self.handle, null);
    }

    pub const ResetError = DeviceDispatch.ResetFencesError;
    pub fn reset(self: *const Self, device: *const Device) ResetError!void {
        try device.dispatch.resetFences(device.handle, 1, @ptrCast([*]const vk.Fence, &self.handle));
    }

    pub const WaitError = DeviceDispatch.WaitForFencesError || DeviceDispatch.GetFenceStatusError;
    pub fn wait(self: *const Self, device: *const Device) WaitError!void {
        const timeout = std.math.maxInt(u64);
        while ((try self.stat(device)) != .signaled) {
            _ = try device.dispatch.waitForFences(device.handle, 1, @ptrCast([*]const vk.Fence, &self.handle), vk.FALSE, timeout);
        }
    }

    pub const StatError = DeviceDispatch.GetFenceStatusError;
    pub fn stat(self: *const Self, device: *const Device) StatError!State {
        return @intToEnum(State, @enumToInt(try device.dispatch.getFenceStatus(device.handle, self.handle)));
    }
};
